/**
***     Copyright (C) 2008 - 2011 Sophia Cradle Incorporated.
***     All rights reserved.
***
***     This source code was written by YAMADA Naoaki.
***
***     Project         : SophiaFramework
***     Codename        : Graviton
***     Platform        : x86, ARM
***     OS              : BREW 2.0, BREW 2.1, BREW 3.1, BREW 4.0
***     Compiler        : Visual C++ 6.0, Visual C++ .NET 2003, Visual C++ 2005,
***                       ADS 1.2, RVCTB 1.2, RVCTB 3.0,
***                       GNUARM 4.1.1, YAGARTO 4.4.2
***     Package         : Graphical User Interface Package
***     File            : SFZWebBrowserControlEx.h.hpp
**/

#ifndef __SOPHIAFRAMEWORK_SFZWEBBROWSERCONTROLEX_HHPP
#define __SOPHIAFRAMEWORK_SFZWEBBROWSERCONTROLEX_HHPP

#include <SFYResponder/SFYEnvironment.h.hpp>
#include "SFZWebBrowserControlEx.f.hpp"
#include <SFZResponder/SFZControl/SFZBoxControl/SFZSingleTextBoxControl.h.hpp>
#include <SFYResponder/SFYControl/SFYControl.h.hpp>
#include <SFXBuffer/SFXBuffer.h.hpp>
#include <SFBWrapper/SFBHTMLViewer.h.hpp>
#include <SFBWrapper/SFBWeb.h.hpp>
#include <SFBWrapper/SFBWebResp.h.hpp>
#include <SFBWrapper/SFBSSLRootCerts.h.hpp>
#include <SFBWrapper/SFBPeek.h.hpp>
#include <SFXGeneral/SFXCallback/SFXTimer.h.hpp>

class SFZWebBrowserControlEx : public SFYControl {
    SFMSEALRESPONDER(SFZWebBrowserControlEx)
    SFMRESPONDERINSTANTIATETHREE(SFZWebBrowserControlEx, SFYControl, SFYWidget, SFYResponder)
    public:
        enum CodeEnum {
            CODE_TYPE                               = four_char_code('C', 'H', 'T', 'M')
        };
        SFMTYPEDEFTYPE(CodeEnum)
        enum DEFAULT {
            DEFAULT_HISTORY_SIZE                    = 30
        };
        SFMTYPEDEFCLASS(Notify)
        class Notify : public HViewNotify {
            public:
                HVNotifyCode        GetType         (Void) const;
                SFBHTMLViewerPtr    GetHTMLViewer   (Void) const;
            protected:
                                    Notify          (SFBHTMLViewerPtr hv, HVNotifyCode notify = HVN_NONE);
        };
        SFMTYPEDEFCLASS(ResponseNotify)
        class ResponseNotify : public Notify {
            public:
                                    ResponseNotify  (SFBHTMLViewerPtr hv, SInt32 response);
                SFCError            GetErrorCode    (Void) const;
                SInt32              GetResponseCode (Void) const;
        };
        SFMTYPEDEFCLASS(DoneNotify)
        class DoneNotify : public Notify {
            public:
                                    DoneNotify      (SFBHTMLViewerPtr hv, SInt32 done);
                SInt32              GetDoneCode     (Void) const;
                SFCError            GetErrorCode    (Void) const;
                UInt32              GetResponseCode (Void) const;
        };
        SFMTYPEDEFCLASS(JumpNotify)
        class JumpNotify : public Notify {
            public:
                                    JumpNotify      (SFBHTMLViewerPtr hv, SInt32 elementNo, ACharConstPtr url, ACharConstPtr method = null, ACharConstPtr data = null);
                                    JumpNotify      (SFBHTMLViewerPtr hv, ACharConstPtr url, ACharConstPtr method = null, ACharConstPtr data = null);
                SInt32              GetElementNo    (Void) const;
                SFXAnsiStringConst  GetUrl          (Void) const;
                SFXAnsiStringConst  GetMethod       (Void) const;
                SFXAnsiStringConst  GetData         (Void) const;
        };
        SFMTYPEDEFCLASS(FocusNotify)
        class FocusNotify : public Notify {
            public:
                                    FocusNotify     (SFBHTMLViewerPtr hv, SInt32 elementNo, HVFocusType type, SInt32 state);
                SInt32              GetElementNo    (Void) const;
                HVFocusType         GetFocusType    (Void) const;
                SInt32              GetState        (Void) const;
        };
        SFMTYPEDEFSTRUCT(HistoryEntry)
        struct HistoryEntry {
            public:
                SFXAnsiString       url;
                SFXAnsiString       title;
        };
    private:
                SFBHTMLViewerSmp                    _viewer;
                SFBWebSmp                           _web;
                SFBWebRespSmp                       _webResp;
                SFXCallback                         _webCallback;
                SInt32                              _webCode;
                SFXAnsiString                       _optMethod;
                SFBPeekSmp                          _optBody;
                VoidPtr                             _optBodyMem;
                SFBSSLRootCertsSmp                  _optSsl;
                SFXAnsiString                       _authHeader;
                SFXAnsiString                       _authBaseUrl;
                HistoryEntryPtr                     _history;
                SFXAnsiStringPtr                    _stateHistory;
                UInt32                              _resumeProperty;
                SFXTimer                            _timer;
                UInt32                              _msecTimeout;
                Bool                                _fullscreen_edit;
                Bool                                _fepin;
    public:
        static  SFZWebBrowserControlExSmp           NewInstance             (SFCErrorPtr exception = null);
                SFCError                            Navigate                (SFXAnsiStringConstRef url);
                SFCError                            Navigate                (SFXAnsiStringConstRef url, SFXAnsiStringConstRef method);
                SFCError                            Navigate                (SFXAnsiStringConstRef url, SFXAnsiStringConstRef method, SFXAnsiStringConstRef data);
                SFCError                            Navigate                (SFXAnsiStringConstRef url, SFXAnsiStringConstRef method, SFXBufferConstRef data);
                SFCError                            Load                    (SFBSourceSmpConstRef source);
                SFCError                            Load                    (SFBSourceSmpConstRef source, SFXAnsiStringConstRef url);
                SFCError                            Load                    (SFBAStreamSmpConstRef stream);
                SFCError                            Load                    (SFBAStreamSmpConstRef stream, SFXAnsiStringConstRef url);
                SFCError                            Load                    (SFXAnsiStringConstRef data);
                SFCError                            Load                    (SFXAnsiStringConstRef data, SFXAnsiStringConstRef url);
                Void                                Reload                  (Void);
                Void                                Stop                    (Void);
                Bool                                IsBusy                  (Void) const;
                SFCError                            Move                    (UInt32 index = 0);
                SFCError                            Previous                (SInt32 n = 1);
                SFCError                            Next                    (SInt32 n = 1);
                Void                                SetWeb                  (SFBWebSmpConstRef web);
                SFBWebSmpConstRef                   GetWeb                  (Void) const;
                Void                                UnsetWeb                (Void);
                SFBWebRespSmpConstRef               GetWebResp              (Void) const;
                HistoryEntry                        GetHistory              (UInt32 index = 0) const;
                HistoryEntry                        GetCurrentHistory       (Void) const;
                HistoryEntryConstPtr                GetHistoryList          (Void) const;
                UInt32                              GetCurrentHistoryIndex  (Void) const;
                Void                                ClearHistory            (Void);
                SFCError                            SetHistorySize          (UInt32 size);
                UInt32                              GetHistorySize          (Void) const;
                SInt32                              FindElement             (ACharConstPtr element, SInt32 no) const;
                SFXAnsiString                       GetElementText          (SInt32 index) const;
                SFCError                            SetAuthorizedData       (SFXAnsiStringConstRef user, SFXAnsiStringConstRef passwd, SFXAnsiStringConstRef url);
                SFCError                            UnsetAuthorizedData     (Void);
                SFCError                            SetIndent               (SInt32 pixel);
                SInt32                              GetIndent               (Void) const;
                SFCError                            SetScrollbarWidth       (SInt32 pixel);
                SInt32                              GetScrollbarWidth       (Void) const;
                SFCError                            SetParagraphSpacing     (SInt32);
                SInt32                              GetParagraphSpacing     (Void) const;
                SFCError                            SetTextColor            (SFXRGBColorConstRef);
                SFXRGBColor                         GetTextColor            (Void) const;
                SFCError                            SetLinkColor            (SFXRGBColorConstRef);
                SFXRGBColor                         GetLinkColor            (Void) const;
                Void                                SetScrollbarEnable      (Bool);
                Void                                SetInlineEditEnable     (Bool);
                Bool                                IsScrollEnable          (Void) const;
                Bool                                IsInlineEditEnable      (Void) const;
                Void                                SetTimeoutCallback      (SFXCallback::CallbackSPP spp, VoidPtr reference);
                Void                                SetTimeoutMilliSec      (UInt32 msec);
                UInt32                              GetTimeoutMilliSec      (Void) const;
    protected:
        explicit                                    SFZWebBrowserControlEx  (Void) static_throws;
        virtual                                     ~SFZWebBrowserControlEx (Void);
        virtual Void                                HandleBoundGlobal       (SFXRectangleConstRef rectangle);
        virtual Void                                HandleRenderRequest     (SFXGraphicsPtr graphics) const;
        virtual Bool                                ResponseNotifyHandler   (ResponseNotifyConstRef notify);
        virtual Void                                DoneNotifyHandler       (DoneNotifyConstRef notify);
        virtual Void                                JumpNotifyHandler       (JumpNotifyConstRef notify);
        virtual Void                                FocusNotifyHandler      (FocusNotifyConstRef notify);
        XANDLER_DECLARE_BOOLEVENT(NativeHandler)
        XANDLER_DECLARE_BOOLVOID(CtlTabHandler)
        XANDLER_DECLARE_VOIDRESUME(ResumeHandler)
        XANDLER_DECLARE_VOIDSUSPEND(SuspendHandler)
    private:
        static  Void                                WebCallback             (VoidPtr p);
        static  Void                                NotifyHandler           (VoidPtr p, HViewNotify* n);
                SFCError                            MakeHandler             (Void);
                SFCError                            OpenImpl                (SFXAnsiStringConstRef url, SFXAnsiStringConstRef method, SFXBufferConstRef data);
                SFCError                            OpenFileImpl            (SFXAnsiStringConstRef url);
                SFCError                            WebSetup                (SFXAnsiStringConstRef url);
                Void                                ClearHistory            (UInt32 index);
                Void                                ClearWebSubset          (Void);
                Void                                ClearWebRespAndOpts     (Void);
                Void                                RemoveHeader            (SFXAnsiStringConstRef header);
                Void                                MakeAuthHeader          (SFXAnsiStringConstRef auth, UIntN len);
                Void                                ReorderHistory          (Void);
                Void                                RenderSafe              (Void);
                Void                                AdjustViewer            (SFXRectangleConstRef rectangle) const;
        XANDLER_DECLARE_VOIDSTATE(OnState)
        XALLBACK_DECLARE_SFXTIMER(OnTimeOutDefault)
};

#include "SFZWebBrowserControlEx.i.hpp"

#endif // __SOPHIAFRAMEWORK_SFZWEBBROWSERCONTROLEX_HHPP //
