/**
***     Copyright (C) 2008 - 2011 Sophia Cradle Incorporated.
***     All rights reserved.
***
***     This source code was written by YAMADA Naoaki.
***
***     Project         : SophiaFramework
***     Codename        : Graviton
***     Platform        : x86, ARM
***     OS              : BREW 2.0, BREW 2.1, BREW 3.1, BREW 4.0
***     Compiler        : Visual C++ 6.0, Visual C++ .NET 2003, Visual C++ 2005,
***                       ADS 1.2, RVCTB 1.2, RVCTB 3.0,
***                       GNUARM 4.1.1, YAGARTO 4.4.2
***     Package         : Graphical User Interface Package
***     File            : SFZWebBrowserControlEx.i.cpp
**/

#include <SFYResponder/SFYApplication/SFYApplication.h.hpp>
#include "SFZWebBrowserControlEx.h.hpp"
#include <SFXGraphics/SFXGraphics.h.hpp>
#include <SFBWrapper/SFBWebUtil.h.hpp>
#include <SFBWrapper/SFBSourceUtil.h.hpp>
#include <SFBWrapper/SFBShell.h.hpp>
#include <SFBWrapper/SFBTextCtl.h.hpp>
#include <SFXGraphics/SFXColor/SFXBevelColor.h.hpp>
#include <SFXGraphics/SFXShape/SFXMargin.h.hpp>

#define     VIEWER_MARGIN           (3)
#define     NULL_STR                SFXAnsiString::EmptyInstance()
#define     NULL_BUFFER             SFXBuffer::EmptyInstance()
#define     UNKNOWN_URL             "file://localhost/"
#define     ADJUST_WIDTH            (1)
#define     ADJUST_HEIGHT           (1)
#define     STATUS_BAR_HEIGHT       (24)

/*protected */SFZWebBrowserControlEx::SFZWebBrowserControlEx(Void) static_throws
{
    SFXRectangle                        rect;
    SFCError                            error(SFERR_NO_ERROR);

    if (static_try()) {
        SetType(CODE_TYPE);
        static_throw(MakeHandler());
        if (static_try()) {
            _webCode = 0;
            _optBodyMem = null;
            _history = null;
            _stateHistory = null;
            _fullscreen_edit = false;
            _fepin = false;
            SetMaximumValue(DEFAULT_HISTORY_SIZE);
            if ((_viewer = SFBHTMLViewer::NewInstance(&error)) != null) {
                if ((_stateHistory = ::new SFXAnsiString[GetHistorySize()]) != null) {
                    if ((_history = ::new HistoryEntry[GetHistorySize()]) != null) {
                        RenderSafe();
                        _viewer->SetRect(SFXRectangle::EmptyInstance());
                        _viewer->SetProperties(HVP_LOADALLIMAGES | HVP_INVALIDATION | _viewer->GetProperties());
                        _viewer->SetNotifyFn(NotifyHandler, this);
                        _timer.Set(XALLBACK_INTERNAL(OnTimeOutDefault));
                        _msecTimeout = 60000;
                    }
                    else {
                        static_throw(SFERR_NO_MEMORY);
                    }
                }
                else {
                    static_throw(SFERR_NO_MEMORY);
                }
            }
            else {
                static_throw(SFERR_FAILED);
            }
        }
    }
}// SFZWebBrowserControlEx::SFZWebBrowserControlEx //

/*protected virtual */SFZWebBrowserControlEx::~SFZWebBrowserControlEx(Void)
{
    if (_viewer != null) {
        RenderSafe();
        Stop();
        _viewer->Reset();
    }
    ClearWebRespAndOpts();
    if (_history != null) {
        ::delete [] _history;
    }
    if (_stateHistory != null) {
        ::delete [] _stateHistory;
    }
}// SFZWebBrowserControlEx::~SFZWebBrowserControlEx //

/*public static */SFZWebBrowserControlExSmp SFZWebBrowserControlEx::NewInstance(SFCErrorPtr exception)
{
    return static_pointer_cast<SFZWebBrowserControlEx>(Factory(::new SFZWebBrowserControlEx, exception));
}// SFZWebBrowserControlEx::NewInstance //

/*public */SFCError SFZWebBrowserControlEx::Navigate(SFXAnsiStringConstRef url)
{
    return Navigate(url, NULL_STR, NULL_BUFFER);
}// SFZWebBrowserControlEx::Navigate(SFXAnsiStringConstRef url) //

/*public */SFCError SFZWebBrowserControlEx::Navigate(SFXAnsiStringConstRef url, SFXAnsiStringConstRef method)
{
    return Navigate(url, method, NULL_BUFFER);
}// SFZWebBrowserControlEx::Navigate(SFXAnsiStringConstRef url, SFXAnsiStringConstRef method) //

/*public */SFCError SFZWebBrowserControlEx::Navigate(SFXAnsiStringConstRef url, SFXAnsiStringConstRef method, SFXAnsiStringConstRef data)
{
    return Navigate(url, method, SFXBuffer(data));
}// SFZWebBrowserControlEx::Navigate(SFXAnsiStringConstRef url, SFXAnsiStringConstRef method, SFXAnsiStringConstRef data) //

/*public */SFCError SFZWebBrowserControlEx::Navigate(SFXAnsiStringConstRef url, SFXAnsiStringConstRef method, SFXBufferConstRef data)
{
    ReorderHistory();
    return OpenImpl(url, method, data);
}// SFZWebBrowserControlEx::Navigate(SFXAnsiStringConstRef url, SFXAnsiStringConstRef method, SFXBufferConstRef data) //

/*public */SFCError SFZWebBrowserControlEx::Load(SFBSourceSmpConstRef source)
{
    return Load(source, UNKNOWN_URL);
}// SFZWebBrowserControlEx::Load(SFBSourceSmpConstRef source) //

/*public */SFCError SFZWebBrowserControlEx::Load(SFBSourceSmpConstRef source, SFXAnsiStringConstRef url)
{
    SFCError                            error(SFERR_NO_ERROR);

    ReorderHistory();
    if (IsBusy()) {
        Stop();
    }
    if ((error = WebSetup(url)) == SFERR_NO_ERROR) {
        _history[GetCurrentHistoryIndex()].url = url;
        _viewer->SetURL(url);
        _viewer->SetIWeb(_web);
        _viewer->LoadSource(source);
    }
    return error;
}// SFZWebBrowserControlEx::Load(SFBSourceSmpConstRef source, SFXAnsiStringConstRef url) //

/*public */SFCError SFZWebBrowserControlEx::Load(SFBAStreamSmpConstRef stream)
{
    return Load(stream, UNKNOWN_URL);
}// SFZWebBrowserControlEx::Load(SFBAStreamSmpConstRef stream) //

/*public */SFCError SFZWebBrowserControlEx::Load(SFBAStreamSmpConstRef stream, SFXAnsiStringConstRef url)
{
    SFCError                            error(SFERR_NO_ERROR);

    ReorderHistory();
    if (IsBusy()) {
        Stop();
    }
    if ((error = WebSetup(url)) == SFERR_NO_ERROR) {
        _history[GetCurrentHistoryIndex()].url = url;
        _viewer->SetURL(url);
        _viewer->SetIWeb(_web);
        _viewer->LoadStream(stream);
    }
    return error;
}// SFZWebBrowserControlEx::Load(SFBAStreamSmpConstRef stream, SFXAnsiStringConstRef url) //

/*public */SFCError SFZWebBrowserControlEx::Load(SFXAnsiStringConstRef data)
{
    return Load(data, UNKNOWN_URL);
}// SFZWebBrowserControlEx::Load(SFXAnsiStringConstRef data) //

/*public */SFCError SFZWebBrowserControlEx::Load(SFXAnsiStringConstRef data, SFXAnsiStringConstRef url)
{
    SFCError error;

    ReorderHistory();
    if (IsBusy()) {
        Stop();
    }
    if ((error = WebSetup(url)) == SFERR_NO_ERROR) {
        _history[GetCurrentHistoryIndex()].url = url;
        _viewer->SetURL(url);
        _viewer->SetIWeb(_web);
        _viewer->SetData(data);
    }
    return error;
}// SFZWebBrowserControlEx::Load(SFXAnsiStringConstRef data, SFXAnsiStringConstRef url) //

/*public */Void SFZWebBrowserControlEx::Reload(Void)
{
    Move(GetCurrentHistoryIndex());
}// SFZWebBrowserControlEx::Reload //

/*public */Void SFZWebBrowserControlEx::Stop(Void)
{
    _webCallback.Cancel();
    _viewer->Stop();
    return;
}// SFZWebBrowserControlEx::Stop //

/*public */Bool SFZWebBrowserControlEx::IsBusy(Void) const
{
    return _webCallback.IsQueued() || _viewer->IsBusy();
}// SFZWebBrowserControlEx::IsBusy //

/*public */SFCError SFZWebBrowserControlEx::Move(UInt32 index)
{
    SFCError                            error(SFERR_NO_ERROR);

    if (index < GetHistorySize() && !_history[index].url.IsEmpty()) {
        _stateHistory[GetCurrentHistoryIndex()] = _viewer->GetViewState();
        SetCurrentValue(index);
        error = OpenImpl(_history[GetCurrentHistoryIndex()].url, NULL_STR, NULL_BUFFER);
    }
    else {
        error = SFERR_INVALID_PARAM;
    }
    return error;
}// SFZWebBrowserControlEx::Move //

/*public */SFCError SFZWebBrowserControlEx::Previous(SInt32 n)
{
    return Next(-n);
}// SFZWebBrowserControlEx::Previous //

/*public */SFCError SFZWebBrowserControlEx::Next(SInt32 n)
{
    SFCError                            error(SFERR_NO_ERROR);

    if (GetCurrentValue() - n >= 0 && GetCurrentValue() - n < GetMaximumValue()) {
        error = Move(static_cast<UInt32>(GetCurrentValue() - n));
    }
    else {
        error = SFERR_INVALID_PARAM;
    }
    return error;
}// SFZWebBrowserControlEx::Next //

/*public */Void SFZWebBrowserControlEx::SetWeb(SFBWebSmpConstRef web)
{
    Stop();
    ClearWebRespAndOpts();
    _web = web;
}// SFZWebBrowserControlEx::SetWeb //

/*public */Void SFZWebBrowserControlEx::UnsetWeb(Void)
{
    Stop();
    ClearWebRespAndOpts();
    _web.Release();
}// SFZWebBrowserControlEx::UnsetWeb //

/*public */SFZWebBrowserControlEx::HistoryEntry SFZWebBrowserControlEx::GetHistory(UInt32 index) const
{
    HistoryEntry dummy;

    if (index < GetHistorySize()) {
        return _history[index];
    }
    else {
        dummy.url = SFXAnsiString::EmptyInstance();
        return dummy;
    }
}// SFZWebBrowserControlEx::GetHistory //

/*public */SFZWebBrowserControlEx::HistoryEntry SFZWebBrowserControlEx::GetCurrentHistory(Void) const
{
    return GetHistory(GetCurrentHistoryIndex());
}// SFZWebBrowserControlEx::GetCurrentHistory //

/*public */SFCError SFZWebBrowserControlEx::SetHistorySize(UInt32 size)
{
    HistoryEntryPtr                     history(null);
    SFXAnsiStringPtr                    stateHistory(null);
    UInt32                              maxSize;
    UInt32                              i;
    SFCError                            error(SFERR_NO_ERROR);

    if (size > GetCurrentHistoryIndex()) {
        if ((history = ::new HistoryEntry[GetHistorySize()]) != null) {
            if ((stateHistory = ::new SFXAnsiString[size]) != null) {
                maxSize = (size < GetHistorySize()) ? (size) : (GetHistorySize());
                for (i = 0; i < maxSize; ++i) {
                    history[i] = _history[i];
                    stateHistory[i] = _stateHistory[i];
                }
                ::delete [] _history;
                ::delete [] _stateHistory;
                SetMaximumValue(size);
                _history = history;
                _stateHistory = stateHistory;
            }
            else {
                ::delete [] history;
                history = null;
                error = SFERR_NO_MEMORY;
            }
        }
        else {
            error = SFERR_NO_MEMORY;
        }
    }
    else {
        error = SFERR_INVALID_PARAM;
    }
    return error;
}// SFZWebBrowserControlEx::SetHistorySize //

/*public */SInt32 SFZWebBrowserControlEx::FindElement(ACharConstPtr element, SInt32 no) const
{
    return _viewer->FindElem(element, no);
}// SFZWebBrowserControlEx::FindElement //

/*public */SFXAnsiString SFZWebBrowserControlEx::GetElementText(SInt32 index) const
{
    SInt32                              len;
    SInt32                              size(128);
    SFXAnsiString                       result;

    for (result.SetThreshold((UInt16)size); (len = _viewer->GetElemText(index, &result)) == size; result.SetThreshold((UInt16)size)) {
        if (len != IHTMLVIEWER_BADELEMENT) {
            size *= 2;
        }
        else {
            result = SFXAnsiString::EmptyInstance();
            break;
        }
    }
    return result;
}// SFZWebBrowserControlEx::GetElemText //

/*public */SFCError SFZWebBrowserControlEx::SetAuthorizedData(SFXAnsiStringConstRef user, SFXAnsiStringConstRef passwd, SFXAnsiStringConstRef url)
{
    SFCError                            error(SFERR_NO_ERROR);

    RemoveHeader(_authHeader);
    _authHeader.Clear();
    if (!user.IsEmpty()) {
        _authBaseUrl = url;
        MakeAuthHeader((user + ":" + passwd).GetCString(), user.GetLength() + passwd.GetLength() + 1);
        if (_authHeader.IsEmpty()) {
            error = SFERR_NO_MEMORY;
        }
    }
    else {
        _authBaseUrl = SFXAnsiString::EmptyInstance();
        error = SFERR_NO_ERROR;
    }
    return error;
}// SFZWebBrowserControlEx::SetAuthorizedData //

/*public */SFCError SFZWebBrowserControlEx::UnsetAuthorizedData(Void)
{
    return SetAuthorizedData(NULL_STR, NULL_STR, NULL_STR);
}// SFZWebBrowserControlEx::UnsetAuthorizedData //

/*public */Void SFZWebBrowserControlEx::ClearHistory(Void)
{
    UInt32                              i;

    for (i = 0; i < GetHistorySize(); ++i) {
        _history[i].title = SFXAnsiString::EmptyInstance();
        _history[i].url = SFXAnsiString::EmptyInstance();
    }
    SetCurrentValue(0);
    return;
}// SFZWebBrowserControlEx::ClearHistory //


/*public */SFCError SFZWebBrowserControlEx::SetIndent(SInt32 pixel)
{
    SFCError                            error(SFERR_NO_ERROR);

    RenderSafe();
    error = _viewer->SetOption(HVOPT_N_INDENT, (HVOpt)pixel);
    Invalidate();
    return error;
}// SFZWebBrowserControlEx::SetIndent //

/*public */SInt32 SFZWebBrowserControlEx::GetIndent(Void) const
{
    HVOpt                               opt;
    SInt32                              pixel;

    if (_viewer->GetOption(HVOPT_N_INDENT, &opt) == SFERR_NO_ERROR) {
        pixel = SInt32(opt);
    }
    else {
        pixel = -1;
    }
    return pixel;
}// SFZWebBrowserControlEx::GetIndent //

/*public */SFCError SFZWebBrowserControlEx::SetScrollbarWidth(SInt32 pixel)
{
    SFCError                            error(SFERR_NO_ERROR);

    RenderSafe();
    error = _viewer->SetOption(HVOPT_N_SBWIDTH, (HVOpt)pixel);
    Invalidate();
    return error;
}// SFZWebBrowserControlEx::SetScrollbarWidth //

/*public */SInt32 SFZWebBrowserControlEx::GetScrollbarWidth(Void) const
{
    HVOpt                               opt;
    SInt32                              pixel;

    if (_viewer->GetOption(HVOPT_N_SBWIDTH, &opt) == SFERR_NO_ERROR) {
        pixel = SInt32(opt);
    }
    else {
        pixel = -1;
    }
    return pixel;
}// SFZWebBrowserControlEx::GetScrollbarWidth //

/*public */SFCError SFZWebBrowserControlEx::SetParagraphSpacing(SInt32 pixel)
{
    SFCError                            error(SFERR_NO_ERROR);

    RenderSafe();
    error = _viewer->SetOption(HVOPT_N_VSPACE, (HVOpt)pixel);
    Invalidate();
    return error;
}// SFZWebBrowserControlEx::SetParagraphSpacing //

/*public */SInt32 SFZWebBrowserControlEx::GetParagraphSpacing(Void) const
{
    HVOpt                               opt;
    SInt32                              pixel;

    if (_viewer->GetOption(HVOPT_N_VSPACE, &opt) == SFERR_NO_ERROR) {
        pixel = SInt32(opt);
    }
    else {
        pixel = -1;
    }
    return pixel;
}// SFZWebBrowserControlEx::GetParagraphSpacing //

/*public */SFCError SFZWebBrowserControlEx::SetTextColor(SFXRGBColorConstRef color)
{
    SFCError                            error(SFERR_NO_ERROR);

    RenderSafe();
    error = _viewer->SetOption(HVOPT_RGB_TEXT, HVOpt(color.AsRGBVAL()));
    Invalidate();
    return error;
}// SFZWebBrowserControlEx::SetTextColor //

/*public */SFXRGBColor SFZWebBrowserControlEx::GetTextColor(Void) const
{
    HVOpt                               opt;
    SFXRGBColor                         color;

    if (_viewer->GetOption(HVOPT_RGB_TEXT, &opt) == SFERR_NO_ERROR) {
        color = SFXRGBColor(RGBVAL(opt));
    }
    return color;
}// SFZWebBrowserControlEx::GetTextColor //

/*public */SFCError SFZWebBrowserControlEx::SetLinkColor(SFXRGBColorConstRef color)
{
    SFCError                            error(SFERR_NO_ERROR);

    RenderSafe();
    error = _viewer->SetOption(HVOPT_RGB_LINK, HVOpt(color.AsRGBVAL()));
    Invalidate();
    return error;
}// SFZWebBrowserControlEx::SetLinkColor //

/*public */SFXRGBColor SFZWebBrowserControlEx::GetLinkColor(Void) const
{
    HVOpt                               opt;
    SFXRGBColor                         color;

    if (_viewer->GetOption(HVOPT_RGB_LINK, &opt) == SFERR_NO_ERROR) {
        color = SFXRGBColor(RGBVAL(opt));
    }
    return color;
}// SFZWebBrowserControlEx::GetLinkColor //

/*public */Void SFZWebBrowserControlEx::SetScrollbarEnable(Bool enable)
{
    UInt32                              property;

    property = _viewer->GetProperties();
    if (enable) {
        property |= HVP_SCROLLBAR;
    }
    else {
        property &= ~HVP_SCROLLBAR;
    }
    RenderSafe();
    _viewer->SetProperties(property);
    Invalidate();
    return;
}// SFZWebBrowserControlEx::SetScrollbarEnable //

/*public */Void SFZWebBrowserControlEx::SetInlineEditEnable(Bool enable)
{
    UInt32                              property;

    property = _viewer->GetProperties();
    if (enable) {
        property |= HVP_INLINE_EDIT;
    }
    else {
        property &= ~HVP_INLINE_EDIT;
    }
    RenderSafe();
    _viewer->SetProperties(property);
    Invalidate();
    return;
}// SFZWebBrowserControlEx::SetInlineEditEnable //

/*public */Bool SFZWebBrowserControlEx::IsScrollEnable(Void) const
{
    return ((_viewer->GetProperties() & HVP_SCROLLBAR) != 0);
}// SFZWebBrowserControlEx::IsScrollEnable //

/*public */Bool SFZWebBrowserControlEx::IsInlineEditEnable(Void) const
{
    return ((_viewer->GetProperties() & HVP_INLINE_EDIT) != 0);
}// SFZWebBrowserControlEx::IsInlineEditEnable //

/*protected virtual */Bool SFZWebBrowserControlEx::ResponseNotifyHandler(ResponseNotifyConstRef notify)
{
    unused(notify);
    return false;
}// SFZWebBrowserControlEx::ResponseNotifyHandler //

/*protected virtual */Void SFZWebBrowserControlEx::DoneNotifyHandler(DoneNotifyConstRef notify)
{
    unused(notify);
    return;
}// SFZWebBrowserControlEx::DoneNotifyHandler //

/*protected virtual */Void SFZWebBrowserControlEx::JumpNotifyHandler(JumpNotifyConstRef notify)
{
    SFXAnsiString                       url;
    SFBWebUtilSmp                       webUtil;
    SFCError                            error(SFERR_NO_ERROR);

    if ((webUtil = SFBWebUtil::NewInstance(&error)) != null) {
        if (!_history[GetCurrentHistoryIndex()].url.IsEmpty()) {
            url = webUtil->MakeUrl(_history[GetCurrentHistoryIndex()].url, notify.GetUrl());
        }
        else {
            url = notify.GetUrl();
        }
        if (notify.GetType() == HVN_JUMP) {
            Navigate(url);
        }
        else {
            if (notify.GetMethod().Compare("POST") == 0) {
                Navigate(url, notify.GetMethod(), notify.GetData());
            }
            else {
                Navigate(url, notify.GetMethod());
            }
        }
    }
    return;
}// SFZWebBrowserControlEx::JumpNotifyHandler //

/*protected virtual */Void SFZWebBrowserControlEx::FocusNotifyHandler(FocusNotifyConstRef notify)
{
    unused(notify);
    return;
}// SFZWebBrowserControlEx::FocusNotifyHandler //

/*private */Void SFZWebBrowserControlEx::RenderSafe(Void)
{
    SFXGraphics::GetInstance()->SetClip(SFXRectangle::EmptyInstance());
}// SFZWebBrowserControlEx::RenderSafe //

/*protected */XANDLER_IMPLEMENT_VOIDSUSPEND(SFZWebBrowserControlEx, SuspendHandler, invoker, reason, info)
{
    unused(invoker); unused(reason); unused(info);
    RenderSafe();
    Stop();
    ClearWebRespAndOpts();
    _web.Release();
    _resumeProperty = _viewer->GetProperties();
    _viewer->Reset();
    return;
}// XANDLER_IMPLEMENT_VOIDSUSPEND //

/*protected */XANDLER_IMPLEMENT_VOIDRESUME(SFZWebBrowserControlEx, ResumeHandler, invoker, environment)
{
    SFCError                            error(SFERR_NO_ERROR);

    unused(invoker); unused(environment);
    RenderSafe();
    if ((_viewer = SFBHTMLViewer::NewInstance(&error)) != null) {
        RenderSafe();
        AdjustViewer(GetGlobalBound());
        _viewer->SetProperties(_resumeProperty);
        _viewer->SetNotifyFn(NotifyHandler, this);
    }
    _viewer->SetActive(GetStateVisible() && GetStateActive() && GetStateEnable());
    _viewer->SetViewState(_stateHistory[GetCurrentHistoryIndex()]);
    Reload();
    return;
}// XANDLER_IMPLEMENT_VOIDRESUME //

/*protected */XANDLER_IMPLEMENT_BOOLVOID(SFZWebBrowserControlEx, CtlTabHandler, invoker)
{
    unused(invoker);
    return false;
}// XANDLER_IMPLEMENT_BOOLVOID(SFZWebBrowserControlEx, CtlTabHandler, invoker)

/*private */SFCError SFZWebBrowserControlEx::OpenImpl(SFXAnsiStringConstRef url, SFXAnsiStringConstRef method, SFXBufferConstRef data)
{
    WebOpt                              webopt[4];
    UIntN                               num(0);
    SFBSourceUtilSmp                    su;
    SFCError                            error(SFERR_NO_ERROR);

    if (url.StartsWith("http://") || url.StartsWith("https://")) {
        if (IsBusy()) {
            Stop();
        }
        if ((error = WebSetup(url)) == SFERR_NO_ERROR) {

            // set additonal http option
            if (!method.IsEmpty()) {
                _optMethod = method;
                webopt[num].nId = WEBOPT_METHOD;
                webopt[num].pVal = _optMethod.GetBuffer();
                ++num;
            }
            if (data.GetSize() != 0) {
                if ((_optBodyMem = MemoryAllocate(data.GetSize())) != null) {
                    SFXHelper::memcpy(_optBodyMem, data.GetBuffer(), data.GetSize());
                    _optBody.Release();
                    if ((su = SFBSourceUtil::NewInstance(&error)) != null) {
                        su->PeekFromMemory(_optBodyMem, data.GetSize(), null, null, &_optBody);
                    }
                    // *** vice versa *** //
                    if (_optBody != null) {
                        webopt[num].nId = WEBOPT_BODY;
                        webopt[num].pVal = VoidPtr(&_optBody);
                        ++num;
                    }
                    else {
                        error = SFERR_FAILED;
                    }
                }
                else {
                    error = SFERR_NO_MEMORY;
                }
            }
            if (error == SFERR_NO_ERROR) {
                webopt[num].nId = WEBOPT_END;
                ASSERT(num < lengthof(webopt) - 1);
                _webCallback.Set(WebCallback, this);
                _history[GetCurrentHistoryIndex()].url = url;
                if ((error = _timer.Schedule(_msecTimeout)) == SFERR_NO_ERROR) {
                    _web->GetResponseV(&_webResp, &_webCallback, _history[GetCurrentHistoryIndex()].url, webopt);
                }
            }
        }
    }
    else {
        error = OpenFileImpl(url);
    }
    return error;
}// SFZWebBrowserControlEx::OpenImpl //

/*private */SFCError SFZWebBrowserControlEx::OpenFileImpl(SFXAnsiStringConstRef url)
{
    static ACharConst                   fileScheme[] = "file://localhost/";
    static ACharConst                   resScheme[] = "res:";
    SFXAnsiString                       path;
    SFBFileMgrSmp                       mgr;
    SFBAStreamSmp                       file;
    SFCError                            error(SFERR_NO_ERROR);

    if (url.StartsWith(fileScheme)) {
        path = url.Substring(STRLEN(fileScheme), url.GetLength());
    }
    else if (url.StartsWith(resScheme)) {
        path = url.Substring(STRLEN(resScheme), url.GetLength());
    }
    else {
        error = SFERR_INVALID_PARAM;
    }
    if (error == SFERR_NO_ERROR) {
        if ((mgr = SFBFileMgr::NewInstance(&error)) != null) {
            file = mgr->OpenFile(path, _OFM_READ);
            if (file != null) {
                if (IsBusy()) {
                    Stop();
                }
                if ((error = WebSetup(url)) == SFERR_NO_ERROR) {
                    _history[GetCurrentHistoryIndex()].url = url;
                    _viewer->SetURL(url);
                    _viewer->SetIWeb(_web);
                    _viewer->LoadStream(file);
                }
            }
            else {
                error = SFERR_INVALID_PARAM; // XXX or mgr->GetLastError();
            }
        }
    }
    return error;
}// SFZWebBrowserControlEx::OpenFileImpl //

/*private static*/Void SFZWebBrowserControlEx::WebCallback(VoidPtr p)
{
    SFZWebBrowserControlExPtr           self(reinterpret_cast<SFZWebBrowserControlExPtr>(p));
    WebRespInfo*                        info = self->_webResp->GetInfo();
    ResponseNotify                      response(self->_viewer.Get(), info->nCode);
    SFBSourceSmp                        source(interface_cast(info->pisMessage));

    self->_timer.Cancel();
    if (!self->ResponseNotifyHandler(response)) {
        self->_webCode = info->nCode;
        if (self->_webCode >= 0) {
            self->_viewer->SetIWeb(self->_web);
            self->_viewer->SetURL(self->_history[self->GetCurrentHistoryIndex()].url);
            self->_viewer->LoadSource(source);
            self->ClearWebRespAndOpts();
        }
        else {
            DoneNotify notify(self->_viewer.Get(), info->nCode);
            NotifyHandler(self, &notify);
        }
    }
}// SFZWebBrowserControlEx::WebCallback //

/*private static */Void SFZWebBrowserControlEx::NotifyHandler(VoidPtr p, HViewNotify* n)
{
    SFZWebBrowserControlExPtr           self(reinterpret_cast<SFZWebBrowserControlExPtr>(p));
    Notify*                             notify = static_cast<Notify*>(n);
    SFBShellSmp                         shell;
    SFXAnsiString                       state;
    SInt32                              elementNo;
    SFYResponderSmp                     root;

    switch (notify->GetType()) {
    case HVN_DONE:
        if (notify->u.nDone == SFERR_NO_ERROR) {
            if ((state = self->_stateHistory[self->GetCurrentHistoryIndex()]) != SFXAnsiString::EmptyInstance()) {
                self->_viewer->SetViewState(state);
            }
            elementNo = self->FindElement("title", 0);
            self->_history[self->GetCurrentHistoryIndex()].title = elementNo < 0 ? SFXAnsiString::EmptyInstance() : self->GetElementText(elementNo);
        }
        if (notify->u.nDone > SFERR_NO_ERROR) {
            self->_viewer->GetOption(HVOPT_N_ERROR, (HVOpt*)(&notify->u.nDone));
            notify->u.nDone = -notify->u.nDone;  // error code change to negative
        }
        else {
            notify->u.nDone = self->_webCode;
        }
        self->DoneNotifyHandler(*(static_cast<DoneNotify*>(notify)));
        break;
    case HVN_JUMP:
    case HVN_SUBMIT:
        self->JumpNotifyHandler(*(static_cast<JumpNotify*>(notify)));
        break;
    case HVN_FOCUS:
        self->FocusNotifyHandler(*(static_cast<FocusNotify*>(notify)));
        break;
    case HVN_FULLSCREEN_EDIT:   // htmlviewer edit field going into fullscreen mode");
        self->_fullscreen_edit = true;
        self->Invalidate();
        break;
    case HVN_REDRAW_SCREEN:     // htmlviewer drew outside of bounding rectangle, applet must redraw it's screen
        self->_fullscreen_edit = false;
        if ((root = self->GetRoot()) != null) {
            root->Render(true);
        }
        break;
    case HVN_INVALIDATE:        // htmlviewer needs redraw
        if (self->_fepin == false && self->GetStateFocus(true)) {
            self->Invalidate();
        }
        else {
            self->_viewer->Redraw();
        }
        break;
    #if TARGET_VERSION_GE(3, 1, 0)
    case HVN_PAGEDONE:
        break;
    case HVN_CONTENTDONE:
        self->Render();
        break;
    #endif
    default:
        break;
    }
}// SFZWebBrowserControlEx::NotifyHandler //

/*private */Void SFZWebBrowserControlEx::ReorderHistory(Void)
{
    UInt32                              i;
    UInt32                              j;

    _stateHistory[GetCurrentHistoryIndex()] = _viewer->GetViewState();
    if (GetCurrentHistoryIndex() == 0) {
        for (i = GetHistorySize() - 1; i >= 1; --i) {
            _history[i] = _history[i - 1];
            _stateHistory[i] = _stateHistory[i - 1];
        }
    }
    else {
        for (i = 1; i < GetHistorySize(); ++i) {
            j = GetCurrentHistoryIndex() + i - 1;
            if (j < GetHistorySize()) {
                _history[i] = _history[j];
                _stateHistory[i] = _stateHistory[j];
            }
            else {
                ClearHistory(i);
            }
        }
    }
    ClearHistory(0);
    SetCurrentValue(0);
}// SFZWebBrowserControlEx::ReorderHistory //

/*private */Void SFZWebBrowserControlEx::ClearHistory(UInt32 index)
{
    _history[index].url = SFXAnsiString::EmptyInstance();
    _history[index].title = SFXAnsiString::EmptyInstance();
    _stateHistory[index] = SFXAnsiString::EmptyInstance();
}// SFZWebBrowserControlEx::ClearHistory //

/*private */Void SFZWebBrowserControlEx::ClearWebSubset(Void)
{
    _webResp.Release();
    _optMethod = SFXAnsiString::EmptyInstance();
    _optBody.Release();
    if (_optBodyMem != null) {
        MemoryFree(_optBodyMem);
    }
    _optBodyMem = null;
    _optSsl.Release();
}// SFZWebBrowserControlEx::ClearWebSubset //

/*private */SFCError SFZWebBrowserControlEx::WebSetup(SFXAnsiStringConstRef url)
{
    SFBWebUtilSmp                       webUtil;
    UrlParts                            parts;
    WebOpt                              webopt[4];
    UIntN                               num;
    SFCError                            error;

    if ((webUtil = SFBWebUtil::NewInstance(&error)) != null) {
        ClearWebSubset();
        if (_web == null) {
            if ((_web = SFBWeb::NewInstance(&error)) == null) {
                error = SFERR_FAILED;
            }
        }
        if (error == SFERR_NO_ERROR) {
            if (webUtil->ParseUrl(url, &parts) == SFERR_NO_ERROR) {
                if (url.StartsWith("http://") || url.StartsWith("https://") || url.StartsWith("file://") || url.StartsWith("res:")) {
                    num = 0;
                    webopt[num].nId = WEBOPT_FLAGS;
                    webopt[num].pVal = null;
                    _web->GetOpt(WEBOPT_FLAGS, null, &webopt[num]);
                    webopt[num].pVal = (VoidPtr)(WEBREQUEST_NOWAITCONN | ((UInt32)webopt[num].pVal));
                    ++num;
                    if (url.StartsWith("https://")) {
                        if ((_optSsl = SFBSSLRootCerts::NewInstance(&error)) != null) {
                            webopt[num].nId = WEBOPT_DEFAULTS;
                            webopt[num].pVal = interface_cast(_optSsl.Get());
                            ++num;
                        }
                        else {
                            error = SFERR_INVALID_PARAM;
                        }
                    }
                    if (error == SFERR_NO_ERROR) {
                        if (!_authHeader.IsEmpty()) {
                            RemoveHeader(_authHeader);
                        }
                        if (UP_HASUSER(&parts)) {
                            _authBaseUrl = SFXAnsiString::EmptyInstance();
                            if (!_authHeader.IsEmpty()) {
                                _authHeader.Clear();
                            }
                            MakeAuthHeader(parts.cpcUser, parts.cpcHost - parts.cpcUser - 1); // -1 for '@'
                            if (_authHeader.IsEmpty()) {
                                error = SFERR_NO_MEMORY;
                            }
                        }
                        else if (!_authBaseUrl.IsEmpty() && (!url.StartsWith(_authBaseUrl))) {
                            if (!_authHeader.IsEmpty()) {
                                _authHeader.Clear();
                            }
                            _authBaseUrl = SFXAnsiString::EmptyInstance();
                        }
                        if (error == SFERR_NO_ERROR) {
                            if (!_authHeader.IsEmpty()) {
                                webopt[num].nId = WEBOPT_HEADER;
                                webopt[num].pVal = _authHeader.GetBuffer();
                                ++num;
                            }
                            webopt[num].nId = WEBOPT_END;
                            ASSERT(num < lengthof(webopt) - 1);
                            error = _web->AddOpt(webopt);
                        }
                    }
                }
                else {
                    error = SFERR_INVALID_PARAM;
                }
            }
            else {
                error = SFERR_INVALID_PARAM;
            }
        }
    }
    return error;
}// SFZWebBrowserControlEx::WebSetup //


/*private */Void SFZWebBrowserControlEx::RemoveHeader(SFXAnsiStringConstRef header)
{
    WebOpt                              webopt;
    SInt32                              n;

    for (n = 0; _web->GetOpt(WEBOPT_HEADER, n, &webopt) == SFERR_NO_ERROR; ++n) {
        if (header.Compare(static_cast<ACharConstPtr>(webopt.pVal)) == 0) {
            _web->RemoveOpt(WEBOPT_HEADER, n);
            break;
        }
    }
}// SFZWebBrowserControlEx::RemoveHeader //

/*private */Void SFZWebBrowserControlEx::MakeAuthHeader(SFXAnsiStringConstRef auth, UIntN len)
{
    SFXAnsiString                       base64;
    SFBWebUtilSmp                       webUtil;
    SFCError                            error;

    base64 = "";
    base64.SetThreshold(BASE64LEN(len) + 1);
    if ((webUtil = SFBWebUtil::NewInstance(&error)) != null) {
        webUtil->EncodeBase64(auth.GetCString(), len, base64.GetBuffer());
        base64.SetChar(BASE64LEN(len), '\0');
    }
    _authHeader = SFXAnsiString::Format("Authorization: Basic %s\r\n", base64.GetCString());
    return;
}// SFZWebBrowserControlEx::MakeAuthHeader //

/*private */Void SFZWebBrowserControlEx::ClearWebRespAndOpts(Void)
{
    _webResp.Release();
    _optMethod = SFXAnsiString::EmptyInstance();
    _optBody.Release();
    if (_optBodyMem != null) {
        MemoryFree(_optBodyMem);
    }
    _optBodyMem = null;
    _optSsl.Release();
}

/*protected virtual */Void SFZWebBrowserControlEx::HandleBoundGlobal(SFXRectangleConstRef rectangle)
{
    AdjustViewer(rectangle);
}// SFZWebBrowserControlEx::HandleBoundGlobal //

/*protected virtual */Void SFZWebBrowserControlEx::HandleRenderRequest(SFXGraphicsPtr graphics) const
{
    static SFXBevelColor::AtomRecConst  frameRgb = {
        {{{0x00, 0xAA, 0xAA, 0xAA}}}, {{{0x00, 0xFF, 0xFF, 0xFF}}}, {{{0x00, 0xFF, 0xFF, 0xFF}}}
    };
    static SFXRGBColor::AtomRecConst    theme[] = {
        {{{0x00, 0x00, 0x00, 0x00}}},
        {{{0x00, 0xDD, 0x99, 0x99}}},
        {{{0x00, 0x99, 0xDD, 0x99}}},
        {{{0x00, 0xFF, 0xFF, 0xFF}}},
        {{{0x00, 0x66, 0x66, 0x66}}}
    };
    SFBDisplaySmp                       display;
    SFXRectangle                        rectangle(GetLocalBound());

    graphics->DrawBevelRectangle(rectangle, frameRgb);
    rectangle.Deflate(1, 1);
    graphics->DrawRectangle(rectangle, theme[0]);
    rectangle.Deflate(1, 1);
    if (GetStateFocus()) {
        graphics->DrawRectangle(rectangle, theme[1]);
    }
    else {
        graphics->DrawRectangle(rectangle, theme[3]);
    }
    if (_fullscreen_edit) {
        display = SFXGraphics::GetInstance()->GetSFBDisplay();
        display->ResetClipRect();
    }
    _viewer->Redraw();
}// SFZWebBrowserControlEx::HandleRenderRequest //

/*protected */XANDLER_IMPLEMENT_BOOLEVENT(SFZWebBrowserControlEx, NativeHandler, invoker, event)
{
    Bool                                result(false);
    unused(invoker);

    if (!_fullscreen_edit) {
        RenderSafe();
    }
    if (_viewer != null && _viewer->IsActive()) {
        if (_viewer->HandleEvent(event)) {
            result = true;
        }
        else {
            #if defined TARGET_EXTENSION_KDDI
            #if TARGET_VERSION_GE(3, 1, 0)
            if (event.GetType() == SFEVT_CTL_TEXT_MODECHANGED) {
                switch (event.GetP16()) {
                    case TMJ_FEPIN://TMJ_FEPIN:
                        _fepin = true;
                        break;
                    case TMJ_FEPOUT://TMJ_FEPOUT:
                        _fepin = false;
                        break;
                    default:
                        break;
                }
            }
            #endif
            #endif
        }
    }
    return result;
}// XANDLER_IMPLEMENT_BOOLEVENT(SFZWebBrowserControlEx, NativeHandler, invoker, event) //

/*private */XANDLER_IMPLEMENT_VOIDSTATE(SFZWebBrowserControlEx, OnState, invoker, reason, state)
{
    Bool                                focus(GetStateFocus(true));
    unused(invoker);
    unused(reason);
    unused(state);

    if (focus != _viewer->IsActive()) {
        _viewer->SetActive(focus);
    }
    return;
}// XANDLER_IMPLEMENT_VOIDSTATE(SFZWebBrowserControlEx, OnState, invoker, reason, state)

/*private */SFCError SFZWebBrowserControlEx::MakeHandler(Void)
{
    static SFXEventRange::AtomRecConst  range[] = {
        {   SFEVT_BEGIN,            SFEVT_END,              SFP16_BEGIN,            SFP16_END           },
        {   SFEVT_CTL_TAB,          SFEVT_CTL_TAB,          SFP16_BEGIN,            SFP16_END           },
        {   SFEVT_APP_RESUME,       SFEVT_APP_RESUME,       SFP16_BEGIN,            SFP16_END           },
        {   SFEVT_APP_SUSPEND,      SFEVT_APP_SUSPEND,      SFP16_BEGIN,            SFP16_END           },
        {   SFEVT_RESPONDER_STATE,  SFEVT_RESPONDER_STATE,  SFP16_STATE_FOCUS,      SFP16_STATE_FOCUS   },
    };
    SFYHandler::RuleRec                 rule[lengthof(range)];
    SFCError                            error(SFERR_NO_ERROR);

    rule[0].spp = XANDLER_FUNCTION(NativeHandler);
    rule[0].reference = this;
    rule[1].spp = XANDLER_FUNCTION(CtlTabHandler);
    rule[1].reference = this;
    rule[2].spp = XANDLER_FUNCTION(ResumeHandler);
    rule[2].reference = this;
    rule[3].spp = XANDLER_FUNCTION(SuspendHandler);
    rule[3].reference = this;
    rule[4].spp = XANDLER_FUNCTION(OnState);
    rule[4].reference = this;
    error = RegisterHandler(atomic_cast(range), rule, lengthof(range));
    return error;
}// SFZWebBrowserControlEx::MakeHandler //

/*private */Void SFZWebBrowserControlEx::AdjustViewer(SFXRectangleConstRef rectangle) const
{
    SFXRectangle                        grect(rectangle);

    grect.Deflate(VIEWER_MARGIN, VIEWER_MARGIN);
    _viewer->SetRect(grect);
}// SFZWebBrowserControlEx::AdjustViewer //

/*private */XALLBACK_IMPLEMENT_SFXTIMER(SFZWebBrowserControlEx, OnTimeOutDefault)
{
    Stop();
    return;
}// XALLBACK_IMPLEMENT_SFXTIMER(SFZWebBrowserControlEx, OnTimeOutDefault)
