/**
***     Copyright (C) 2008 - 2011 Sophia Cradle Incorporated.
***     All rights reserved.
***
***     This source code was written by YAMADA Naoaki.
***
***     Project         : SophiaFramework
***     Codename        : Graviton
***     Platform        : x86, ARM
***     OS              : BREW 2.0, BREW 2.1, BREW 3.1, BREW 4.0
***     Compiler        : Visual C++ 6.0, Visual C++ .NET 2003, Visual C++ 2005,
***                       ADS 1.2, RVCTB 1.2, RVCTB 3.0,
***                       GNUARM 4.1.1, YAGARTO 4.4.2
***     Package         : Graphical User Interface Package
***     File            : SFZWebBrowserControlEx.i.hpp
**/

#ifndef __SOPHIAFRAMEWORK_SFZWEBBROWSERCONTROLEX_IHPP
#define __SOPHIAFRAMEWORK_SFZWEBBROWSERCONTROLEX_IHPP

#include <SFYResponder/SFYEnvironment.h.hpp>

/*public */inline SFZWebBrowserControlEx::Notify::Notify(SFBHTMLViewerPtr hv, HVNotifyCode notifyCode)
{
    phv = interface_cast(hv);
    code = notifyCode;
}// Notify::Notify //

/*public */inline HVNotifyCode SFZWebBrowserControlEx::Notify::GetType(Void) const
{
    return code;
}// Notify::GetType //

/*public */inline SFBHTMLViewerPtr SFZWebBrowserControlEx::Notify::GetHTMLViewer(Void) const
{
    return interface_cast(phv);
}// Notify::GetHTMLViewer //

/*public */inline SFZWebBrowserControlEx::ResponseNotify::ResponseNotify(SFBHTMLViewerPtr hv, SInt32 response)
    : Notify(hv, HVN_NONE)
{
    u.nDone = response;
}// ResponseNotify:ResponseNotify //

/*public */inline SInt32 SFZWebBrowserControlEx::ResponseNotify::GetErrorCode(Void) const
{
    if (u.nDone >= 0) {
        return SFERR_NO_ERROR;
    }
    return -u.nDone;
}// ResponseNotify::GetErrorCode //

/*public */inline SInt32 SFZWebBrowserControlEx::ResponseNotify::GetResponseCode(Void) const
{
    if (u.nDone < 0) {
        return 0;
    }
    return u.nDone;
}// ResponseNotify::GetResponseCode //

/*public */inline SFZWebBrowserControlEx::DoneNotify::DoneNotify(SFBHTMLViewerPtr hv, SInt32 done)
    : Notify(hv, HVN_DONE)
{
    u.nDone = done;
}// DoneNotify::DoneNotify //

/*public */inline SInt32 SFZWebBrowserControlEx::DoneNotify::GetDoneCode(Void) const
{
    return u.nDone;
}

/*public */inline SInt32 SFZWebBrowserControlEx::DoneNotify::GetErrorCode(Void) const
{
    if (u.nDone >= 0) {
        return SFERR_NO_ERROR;
    }
    return -u.nDone;
}

/*public */inline UInt32 SFZWebBrowserControlEx::DoneNotify::GetResponseCode(Void) const
{
    if (u.nDone < 0) {
        return 0;
    }
    return u.nDone;
}

/*public */inline SFZWebBrowserControlEx::JumpNotify::JumpNotify(SFBHTMLViewerPtr hv, ACharConstPtr url, ACharConstPtr method, ACharConstPtr data)
    : Notify(hv, HVN_JUMP)
{
    JumpNotify(hv, 0, url, method, data);
}

/*public */inline SFZWebBrowserControlEx::JumpNotify::JumpNotify(SFBHTMLViewerPtr hv, SInt32 elementNo, ACharConstPtr url, ACharConstPtr method, ACharConstPtr data)
    : Notify(hv, HVN_JUMP)
{
    u.jump.posElem = elementNo;
    if (method != null) {
        code = HVN_SUBMIT;
        u.jump.pszURL = url;
        u.jump.pszMethod = method;
        u.jump.pszData = data;
    }
}

/*public */inline SInt32 SFZWebBrowserControlEx::JumpNotify::GetElementNo(Void) const
{
    return u.jump.posElem;
}

/*public */inline SFXAnsiStringConst SFZWebBrowserControlEx::JumpNotify::GetUrl(Void) const
{
    return u.jump.pszURL;
}

/*public */inline SFXAnsiStringConst SFZWebBrowserControlEx::JumpNotify::GetMethod(Void) const
{
    return u.jump.pszMethod;
}

/*public */inline SFXAnsiStringConst SFZWebBrowserControlEx::JumpNotify::GetData(Void) const
{
    return u.jump.pszData;
}

/*public */inline SFZWebBrowserControlEx::FocusNotify::FocusNotify(SFBHTMLViewerPtr hv, SInt32 elementNo, HVFocusType type, SInt32 state)
    : Notify(hv, HVN_FOCUS)
{
    u.focus.posElem = elementNo;
    u.focus.nType = type;
    u.focus.nState = state;
}

/*public */inline SInt32 SFZWebBrowserControlEx::FocusNotify::GetElementNo(Void) const
{
    return u.focus.posElem;
}

/*public */inline HVFocusType SFZWebBrowserControlEx::FocusNotify::GetFocusType(Void) const
{
    return u.focus.nType;
}

/*public */inline SInt32 SFZWebBrowserControlEx::FocusNotify::GetState(Void) const
{
    return u.focus.nState;
}

/*public */inline SFBWebSmpConstRef SFZWebBrowserControlEx::GetWeb(Void) const
{
    return _web;
}// SFZWebBrowserControlEx::GetWeb //

/*public */inline SFBWebRespSmpConstRef SFZWebBrowserControlEx::GetWebResp(Void) const
{
    return _webResp;
}// SFZWebBrowserControlEx::GetWebResp //

/*public */inline UInt32 SFZWebBrowserControlEx::GetHistorySize(Void) const
{
    return static_cast<UInt32>(GetMaximumValue());
}

/*public */inline UInt32 SFZWebBrowserControlEx::GetCurrentHistoryIndex(Void) const
{
    return static_cast<UInt32>(GetCurrentValue());
}// SFZWebBrowserControlEx::GetCurrentHistoryIndex //

/*public */inline SFZWebBrowserControlEx::HistoryEntryConstPtr SFZWebBrowserControlEx::GetHistoryList(Void) const
{
    return _history;
}// SFZWebBrowserControlEx::GetHistoryList //

/*public */inline Void SFZWebBrowserControlEx::SetTimeoutCallback(SFXCallback::CallbackSPP spp, VoidPtr reference)
{
    _timer.Set(spp, reference);
    return;
}// SFZWebBrowserControlEx::SetTimeoutCallback //

/*public */inline Void SFZWebBrowserControlEx::SetTimeoutMilliSec(UInt32 msec)
{
    _msecTimeout = msec;
    return;
}// SFZWebBrowserControlEx::SetTimeoutMilliSec //

/*public */inline UInt32 SFZWebBrowserControlEx::GetTimeoutMilliSec(Void) const
{
    return _msecTimeout;
}// SFZWebBrowserControlEx::GetTimeoutMilliSec //

#endif // __SOPHIAFRAMEWORK_SFZWEBBROWSERCONTROLEX_IHPP //
